/**
 * HTTP代理服务器
 * @author Hoverlees http://www.hoverlees.com/blog/?p=1337
 */

#include "../lua/lua.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "proxy_config.h"

static int proxy_check_ip(const char *ip,int len) {
	char number[4];
	int i;
	int j=0;
	int k=0;
	number[0]=number[1]=number[2]=number[3]=0;
	for(i=0;i<len;i++){
		if(ip[i]>='0' && ip[i]<='9'){
			if(j>=3) return 0;
			number[j]=ip[i];
			j++;
		}
		else if(ip[i]=='.'){
			if(j>0){
				if(atoi(number)>255) return 0;
				number[0]=number[1]=number[2]=number[3]=0;
				j=0;
				k++;
			}
			else return 0;
		}
		else return 0;
	}
	if(k!=3) return 0;
	if(atoi(number)>255) return 0;
	return 1;
}

static void* proxy_allocator(void *ud,void *ptr,size_t osize,size_t nsize){
	if(nsize==0){
		if(ptr!=NULL) free(ptr);
		return NULL;
	}
	else{
		if(ptr==NULL) return malloc(nsize);
		else return realloc(ptr,nsize);
	}
}

int proxy_request_callback(ProxyConfig* config,ProxyHttpRequest* request){
	int r=0;
	lua_State *runtime;
	ProxyHttpHeaderNode* node;
	runtime=config->runtime;
	lua_pushstring(runtime,"on_request");
	lua_gettable(runtime,-2);
	lua_newtable(runtime);
	lua_pushstring(runtime,"requestMethod");
	lua_pushstring(runtime,request->method);
	lua_settable(runtime,-3);
	lua_pushstring(runtime,"requestURI");
	lua_pushstring(runtime,request->uri);
	lua_settable(runtime,-3);
	lua_pushstring(runtime,"requestBody");
	lua_pushstring(runtime,request->method);
	lua_settable(runtime,-3);
	node=request->headers;
	while(node){
		lua_pushstring(runtime,node->key);
		lua_pushstring(runtime,node->value);
		lua_settable(runtime,-3);
		node=node->next;
	}
	if(lua_pcall(runtime,1,1,0)==0){
		r=lua_toboolean(runtime,-1);
		lua_pop(runtime,1);
	}
	else{
		fprintf(stderr,"lua error: %s\n",lua_tostring(runtime,-1));
		lua_pop(runtime,1);
	}
	return r;
}

ProxyConfig* proxy_read_config(const char* filename){
	lua_State *runtime;
	ProxyConfig* config;
	const char* t;
	int tlen,i;
	runtime=lua_newstate(proxy_allocator,NULL);
	if(!runtime) return NULL;
	luaL_openlibs(runtime);
	if(luaL_loadfile(runtime,filename)!=0){
		lua_close(runtime);
		printf("load file `%s` error\n",filename);
		return NULL;
	}
	if(lua_pcall(runtime,0,LUA_MULTRET,0)!=0){
		fprintf(stderr,"lua error: %s\n",lua_tostring(runtime,-1));
		lua_pop(runtime,1);
		return NULL;
	}
	lua_getglobal(runtime,"proxy_config");
	if(lua_isnil(runtime,-1)){
		fprintf(stderr,"error: can't find `proxy_config` variable in the config file '%s'.\n",filename);
		lua_close(runtime);
		return NULL;
	}
	if(!lua_istable(runtime,-1)){
		fprintf(stderr,"error: `proxy_config` is not a valid lua table.\n");
		lua_close(runtime);
		return NULL;
	}
	config=(ProxyConfig*) calloc(1,sizeof(ProxyConfig));
	
	lua_pushstring(runtime,"listen");
	lua_gettable(runtime,-2);
	t=lua_tostring(runtime,-1);
	tlen=strlen(t);
	if(proxy_check_ip(t,tlen)) strncpy(config->bindIP,t,31);
	else{
		fprintf(stderr,"warning: invalid ip `%s` using `0.0.0.0`\n",t);
		strcpy(config->bindIP,"0.0.0.0");
	}
	lua_pop(runtime,1);
	
	lua_pushstring(runtime,"port");
	lua_gettable(runtime,-2);
	config->bindPort=lua_tointeger(runtime,-1);
	if(config->bindPort==0) config->bindPort=8766;
	lua_pop(runtime,1);
	
	config->runtime=runtime;
	return config;
}
void proxy_free_config(ProxyConfig* config){
	if(config==NULL) return;
	lua_close(config->runtime);
	free(config);
}