/**
 * HTTP代理服务器
 * @author Hoverlees http://www.hoverlees.com/blog/?p=1337
 */

#include "proxy_epoll_server.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void epoll_send_queue_destory(EpollSendQueue* queue);

void closesocket(SOCKET socket){
	shutdown(socket,SHUT_RDWR);
	close(socket);
}
static void epoll_free_session(EpollSession* session){
	closesocket(session->socket);
	epoll_send_queue_destory(&session->sendQueue);
	free(session);
}
static void epoll_free_send_buffer(EpollSendBuffer* buffer){
	if(buffer->freeFlag&EPOLL_FLAG_FREE_BUFFER) free(buffer->buffer);
	if(buffer->freeFlag&EPOLL_FLAG_FREE_SELF) free(buffer);
}

static int epoll_send_queue_init(EpollSendQueue* queue,int initSize){
	queue->buffers=(EpollSendBuffer**)malloc(sizeof(EpollSendBuffer*)*initSize);
	if(!queue->buffers){
		return 0;
	}
	queue->size=initSize;
	queue->head=0;
	queue->tail=0;
	queue->itemNum=0;
	return 1;
}
static void epoll_send_queue_destory(EpollSendQueue* queue){
	int i,j;
	j=queue->head;
	for(i=0;i<queue->itemNum;i++){
		epoll_free_send_buffer(queue->buffers[j]);
		j++;
		if(j>=queue->size) j=0;
	}
	free(queue->buffers);
}
static int epoll_send_queue_in(EpollSendQueue* queue,EpollSendBuffer* buffer){
	EpollSendBuffer** t;
	if(queue->itemNum>=queue->size){
		return -1;
	}
	queue->buffers[queue->tail]=buffer;
	queue->tail++;
	queue->itemNum++;
	if(queue->tail>=queue->size) queue->tail=0;
	return queue->itemNum;
}
static EpollSendBuffer* epoll_send_queue_head(EpollSendQueue* queue){
	if(queue->itemNum>0) return queue->buffers[queue->head];
	return NULL;
}
static EpollSendBuffer* epoll_send_queue_out(EpollSendQueue* queue){
	EpollSendBuffer* ret;
	if(queue->itemNum==0) return NULL;
	ret=queue->buffers[queue->head];
	queue->itemNum--;
	queue->head++;
	if(queue->head>=queue->size) queue->head=0;
	return ret;
}

int epoll_server_create(EpollServer* server,const char* address,int port){
	int backlog=0;
	memset(server,0,sizeof(EpollServer));
	server->serverSocket=socket(AF_INET,SOCK_STREAM,0);
	if(server->serverSocket<0){
		return 0;
	}
	server->address.sin_family=AF_INET;
	server->address.sin_port=htons(port);
	inet_aton(address,&server->address.sin_addr);
	if(bind(server->serverSocket,(struct sockaddr*)&server->address,sizeof(struct sockaddr_in))!=0){
		fprintf(stderr,"error: can't bind address %s:%d\n",address,port);
		return 0;
	}
	if(listen(server->serverSocket,backlog)!=0){
		fprintf(stderr,"error: can't listen address %s:%d\n",address,port);
		return 0;
	}
	fcntl(server->serverSocket,F_SETFL,fcntl(server->serverSocket,F_GETFL,0)|O_NONBLOCK);
	server->epoll=epoll_create(1);
	if(server->epoll<=0){
		closesocket(server->serverSocket);
		return 0;
	}
	return 1;
}

int epoll_server_start(EpollServer* server){
	struct epoll_event sevent,nevent;
	char buffer[2049];
	int rsize,wsize;
	int opt=1;
	EpollSession* session;
	EpollSendBuffer* sbuffer;
	sevent.data.ptr=NULL;
	sevent.events=EPOLLIN;
	epoll_ctl(server->epoll,EPOLL_CTL_ADD,server->serverSocket,&sevent);
	server->running=1;
	while(server->running){
		if(epoll_wait(server->epoll,&sevent,1,5000)>0){
			if(sevent.data.ptr==NULL){ //accept
				session=(EpollSession*) calloc(sizeof(EpollSession),1);
				if(session){
					rsize=sizeof(struct sockaddr_in);
					session->socket=accept(server->serverSocket,(struct sockaddr*)&session->address,&rsize);
					if(session->socket<0){
						free(session);
						continue;
					}
					if(!epoll_send_queue_init(&session->sendQueue,EPOLL_MAX_SESSION_SEND_BUFFER)){
						closesocket(session->socket);
						free(session);
						continue;
					}
					session->sendPosition=0;
					session->server=server;
					fcntl(session->socket,F_SETFL,fcntl(session->socket,F_GETFL,0)|O_NONBLOCK);
					setsockopt(session->socket,SOL_SOCKET,SO_KEEPALIVE,&opt,sizeof(opt));
					if(server->onConnect) server->onConnect(session);
					nevent.events=EPOLLIN;
					nevent.data.ptr=session;
					epoll_ctl(server->epoll,EPOLL_CTL_ADD,session->socket,&nevent);
				}
				else{
					fprintf(stderr,"can't alloc session,memory full.\n");
					closesocket(accept(server->serverSocket,NULL,NULL));
				}
			}
			else{
				session=(EpollSession*) sevent.data.ptr;
				if(sevent.events&EPOLLIN){
					while(1){
						rsize=recv(session->socket,buffer,2048,0);
						if(rsize>0){
							if(server->onRecieve!=NULL) server->onRecieve(session,buffer,rsize);
						}
						else if(rsize==-1 && (errno==EAGAIN || errno==EWOULDBLOCK)){
							break;
						}
						else{
							if(server->onDisconnect!=NULL) server->onDisconnect(session);
							epoll_ctl(server->epoll,EPOLL_CTL_DEL,session->socket,NULL);
							epoll_free_session(session);
							sevent.events=0;
							break;
						}
					}
				}
				if(sevent.events&EPOLLOUT){
					sbuffer=epoll_send_queue_head(&session->sendQueue);
					while(sbuffer!=NULL){
						if(sbuffer->buffer==NULL){
							nevent.data.ptr=session;
							nevent.events=EPOLLIN;
							epoll_ctl(server->epoll,EPOLL_CTL_MOD,session->socket,&nevent);
							shutdown(session->socket,SHUT_WR);
							epoll_free_send_buffer(sbuffer);
							epoll_send_queue_out(&session->sendQueue);
							break;
						}
						else{
							wsize=sbuffer->length-session->sendPosition;
							rsize=send(session->socket,sbuffer->buffer+session->sendPosition,wsize,0);
							if(rsize>0) session->sendPosition+=rsize;
							if(rsize==wsize){
								session->sendPosition=0;
								if(server->onSent){
									server->onSent(session,sbuffer);
								}
								epoll_free_send_buffer(sbuffer);
								epoll_send_queue_out(&session->sendQueue);
								sbuffer=epoll_send_queue_head(&session->sendQueue);
							}
						}
					}
					if(sbuffer==NULL){
						session->sendPosition=0;
						nevent.data.ptr=session;
						nevent.events=EPOLLIN;
						epoll_ctl(server->epoll,EPOLL_CTL_MOD,session->socket,&nevent);
					}
				}
				if(sevent.events&EPOLLERR){
					if(server->onDisconnect!=NULL) server->onDisconnect(session);
					epoll_free_session(session);
					sevent.events=0;
				}
			}
		}
	}
	return 0;
}

EpollSession* epoll_server_register(EpollServer* server,SOCKET socket){
	struct epoll_event nevent;
	EpollSession* session;
	int opt=1;
	session=(EpollSession*) calloc(sizeof(EpollSession),1);
	if(session){
		session->socket=socket;
		if(!epoll_send_queue_init(&session->sendQueue,EPOLL_MAX_SESSION_SEND_BUFFER)){
			closesocket(session->socket);
			free(session);
			return NULL;
		}
		session->sendPosition=0;
		session->server=server;
		fcntl(session->socket,F_SETFL,fcntl(session->socket,F_GETFL,0)|O_NONBLOCK);
		setsockopt(session->socket,SOL_SOCKET,SO_KEEPALIVE,&opt,sizeof(opt));
		nevent.events=EPOLLIN;
		nevent.data.ptr=session;
		epoll_ctl(server->epoll,EPOLL_CTL_ADD,session->socket,&nevent);
	}
	else{
		fprintf(stderr,"can't alloc client session,memory full.\n");
		return NULL;
	}
	return session;
}

void epoll_server_force_free_session(EpollSession* session){
	struct epoll_event sevent;
	if(session->server->onDisconnect!=NULL) session->server->onDisconnect(session);
	epoll_ctl(session->server->epoll,EPOLL_CTL_DEL,session->socket,NULL);
	epoll_free_session(session);
}

int epoll_server_send(EpollSession* session,EpollSendBuffer* buffer){
	int n;
	struct epoll_event nevent;
	n=epoll_send_queue_in(&session->sendQueue,buffer);
	if(n==1){
		nevent.data.ptr=session;
		nevent.events=EPOLLIN|EPOLLOUT;
		epoll_ctl(session->server->epoll,EPOLL_CTL_MOD,session->socket,&nevent);
	}
	return n;
}

int epoll_server_send_eof(EpollSession* session){
	int n;
	struct epoll_event nevent;
	EpollSendBuffer* buffer=(EpollSendBuffer*)malloc(sizeof(EpollSendBuffer));
	if(buffer==NULL) return 0;
	buffer->freeFlag=EPOLL_FLAG_FREE_SELF;
	buffer->buffer=NULL;
	buffer->length=0;
	n=epoll_send_queue_in(&session->sendQueue,buffer);
	if(n==1){
		nevent.data.ptr=session;
		nevent.events=EPOLLIN|EPOLLOUT;
		epoll_ctl(session->server->epoll,EPOLL_CTL_MOD,session->socket,&nevent);
	}
	return n;
}

int epoll_server_send_data(EpollSession* session,char* data,int len){
	int n;
	struct epoll_event nevent;
	EpollSendBuffer* buffer=(EpollSendBuffer*)malloc(sizeof(EpollSendBuffer));
	if(buffer==NULL) return 0;
	buffer->freeFlag=EPOLL_FLAG_FREE_SELF;
	buffer->buffer=data;
	buffer->length=len;
	n=epoll_send_queue_in(&session->sendQueue,buffer);
	if(n==1){
		nevent.data.ptr=session;
		nevent.events=EPOLLIN|EPOLLOUT;
		epoll_ctl(session->server->epoll,EPOLL_CTL_MOD,session->socket,&nevent);
	}
	return n;
}

int epoll_server_send_copy(EpollSession* session,char* data,int len){
	int n;
	char* t;
	struct epoll_event nevent;
	EpollSendBuffer* buffer=(EpollSendBuffer*)malloc(sizeof(EpollSendBuffer)+len);
	if(buffer==NULL) return 0;
	t=(char*) buffer;
	buffer->freeFlag=EPOLL_FLAG_FREE_SELF;
	buffer->buffer=t+sizeof(EpollSendBuffer);
	buffer->length=len;
	memcpy(buffer->buffer,data,len);
	n=epoll_send_queue_in(&session->sendQueue,buffer);
	if(n==1){
		nevent.data.ptr=session;
		nevent.events=EPOLLIN|EPOLLOUT;
		epoll_ctl(session->server->epoll,EPOLL_CTL_MOD,session->socket,&nevent);
	}
	return n;
}