/**
 * HTTP代理服务器
 * @author Hoverlees http://www.hoverlees.com/blog/?p=1337
 */

#ifndef __HOVERLEES_EPOLL_SERVER_H
#define __HOVERLEES_EPOLL_SERVER_H

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/epoll.h>
#include <pthread.h>
#include <fcntl.h>
#include <errno.h>

#ifndef SOCKET
typedef int SOCKET;
#endif

#define EPOLL_MAX_SESSION_SEND_BUFFER 256

#define EPOLL_FLAG_FREE_BUFFER	1
#define EPOLL_FLAG_FREE_SELF	2

typedef struct{
	int length;
	char* buffer;
	int freeFlag;
}EpollSendBuffer;

typedef struct _EpollSession EpollSession;

typedef void (*_epoll_server_onconnect)(EpollSession* session);
typedef void (*_epoll_server_onrecieve)(EpollSession* session,char* buffer,int size);
typedef void (*_epoll_server_onsent)(EpollSession* session,EpollSendBuffer* buffer);
typedef void (*_epoll_server_ondisconnect)(EpollSession* session);

typedef struct{
	int size;
	int head;
	int tail;
	int itemNum;
	EpollSendBuffer** buffers;
}EpollSendQueue;

typedef struct{
	int running;
	SOCKET serverSocket;
	struct sockaddr_in address;
	int epoll;
	EpollSession* sessions;
	
	_epoll_server_onconnect onConnect;
	_epoll_server_onrecieve onRecieve;
	_epoll_server_onsent onSent;
	_epoll_server_ondisconnect onDisconnect;
}EpollServer;

struct _EpollSession{
	EpollServer* server;
	SOCKET socket;
	struct sockaddr_in address;
	EpollSendQueue sendQueue;
	int sendPosition;
	void* userData;
};

void closesocket(SOCKET socket);

int epoll_server_create(EpollServer* server,const char* address,int port);
int epoll_server_start(EpollServer* server);
EpollSession* epoll_server_register(EpollServer* server,SOCKET socket);
void epoll_server_force_free_session(EpollSession* session);

int epoll_server_send(EpollSession* session,EpollSendBuffer* buffer);
int epoll_server_send_eof(EpollSession* session);
int epoll_server_send_data(EpollSession* session,char* data,int len);
int epoll_server_send_copy(EpollSession* session,char* data,int len);

#endif