/**
 * HTTP代理服务器
 * @author Hoverlees http://www.hoverlees.com/blog/?p=1337
 */

#include "proxy_http_request.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define PROXY_HTTP_REQUEST_PARSE_STATUS_REQUEST	0
#define PROXY_HTTP_REQUEST_PARSE_STATUS_HEADER	1
#define PROXY_HTTP_REQUEST_PARSE_STATUS_VALUE	2
#define PROXY_HTTP_REQUEST_PARSE_STATUS_BODY	3
#define PROXY_HTTP_REQUEST_PARSE_STATUS_FINISH	4

static int proxy_http_request_create_header_node(ProxyHttpRequest* request){
	char* t;
	int keyLen;
	int valueLen;
	ProxyHttpHeaderNode* node;
	keyLen=request->pend1-request->pstart1;
	valueLen=request->pend2-request->pstart2;
	node=(ProxyHttpHeaderNode*) calloc(sizeof(ProxyHttpHeaderNode)+keyLen+valueLen+2,1);
	if(node==NULL) return 0;
	node->next=NULL;
	t=(char*)node;
	t+=sizeof(ProxyHttpHeaderNode);
	node->key=t;
	t=t+1+keyLen;
	node->value=t;
	strncpy(node->key,request->rawRequest+request->pstart1,keyLen);
	strncpy(node->value,request->rawRequest+request->pstart2,valueLen);
	
	if(request->headers==NULL){
		request->headers=node;
		request->lastHeader=node;
	}
	else{
		request->lastHeader->next=node;
		request->lastHeader=node;
	}
	
	return 1;
}

static int proxy_http_request_parse_request(ProxyHttpRequest* request,int start,int end){
	int i;
	int n;
	char* t;
	t=(char*) malloc(end-start);
	if(t==NULL) return 0;
	request->method=t;
	i=start;
	n=0;
	while(i<end){
		if(request->rawRequest[i]==' '){
			*t=0;
			n++;
			if(n>=2) break;
			request->uri=t+1;
		}
		else{
			*t=request->rawRequest[i];
		}
		t++;
		i++;
	}
	return 1;
}

ProxyHttpRequest* proxy_http_request_create(){
	ProxyHttpRequest* ret;
	ret=(ProxyHttpRequest*) calloc(sizeof(ProxyHttpRequest),1);
	if(ret!=NULL){
		ret->status=PROXY_HTTP_REQUEST_PARSE_STATUS_REQUEST;
		ret->rawRequestBufferSize=10240;
		ret->rawRequest=(char*) malloc(10240);
		ret->rawRequestLen=0;
		if(ret->rawRequest==NULL){
			free(ret);
			return NULL;
		}
	}
	return ret;
}

void proxy_http_request_destroy(ProxyHttpRequest* request){
	ProxyHttpHeaderNode *node,*t;
	if(request==NULL) return;
	node=request->headers;
	while(node){
		t=node;
		node=node->next;
		free(t);
	}
	free(request->rawRequest);
	if(request->method!=NULL) free(request->method);
	free(request);
}

char* proxy_http_request_get_header(ProxyHttpRequest* request,const char* key){
	char* r=NULL;
	ProxyHttpHeaderNode* p;
	p=request->headers;
	while(p){
		if(strcasecmp(key,p->key)==0){
			r=p->value;
			break;
		}
		p=p->next;
	}
	return r;
}


int proxy_http_request_add_buffer(ProxyHttpRequest* request,char* buffer,int len){
	char* t;
	char c;
	int i;
	if(request->rawRequestLen+len>=request->rawRequestBufferSize){
		request->rawRequestBufferSize+=10240;
		t=(char*) realloc(request->rawRequest,request->rawRequestBufferSize);
		if(t==NULL) return -1;
		request->rawRequest=t;
	}
	t=request->rawRequest;
	t=t+request->rawRequestLen;
	for(i=0;i<len;i++){
		*t=buffer[i];
		t++;
	}
	request->rawRequestLen+=len;
	while(request->ppos<request->rawRequestLen){
		c=request->rawRequest[request->ppos];
		switch(request->status){
			case PROXY_HTTP_REQUEST_PARSE_STATUS_HEADER:
				if(c==':'){
					request->pstart2=request->ppos+1;
					request->pend2=request->pstart2;
					request->status=PROXY_HTTP_REQUEST_PARSE_STATUS_VALUE;
				}
				else if(c=='\r'){
					request->ppos++;
					request->pstart1=request->ppos+1;
					request->pend1=request->pstart1;
					t=proxy_http_request_get_header(request,"Content-Length");
					if(t==NULL){
						request->status=PROXY_HTTP_REQUEST_PARSE_STATUS_FINISH;
						return 1;
					}
					else{
						request->requestBody=request->rawRequest+request->ppos+1;
						request->requestBodyLen=atoi(t);
						request->status=PROXY_HTTP_REQUEST_PARSE_STATUS_BODY;
					}
				}
				else request->pend1++;
				break;
			case PROXY_HTTP_REQUEST_PARSE_STATUS_VALUE:
				if(c=='\r'){
					if(!proxy_http_request_create_header_node(request)) return -1;
					request->ppos++;
					request->pstart1=request->ppos+1;
					request->pend1=request->pstart1;
					request->status=PROXY_HTTP_REQUEST_PARSE_STATUS_HEADER;
				}
				else if(c==' ' && request->pstart2==request->pend2){
					request->pstart2++;
					request->pend2++;
				}
				else request->pend2++;
				break;
			case PROXY_HTTP_REQUEST_PARSE_STATUS_BODY:
				request->pend1++;
				if(request->pend1-request->pstart1>=request->requestBodyLen){
					request->status=PROXY_HTTP_REQUEST_PARSE_STATUS_FINISH;
					return 1;
				}
				break;
			case PROXY_HTTP_REQUEST_PARSE_STATUS_REQUEST:
				if(c=='\r'){
					if(!proxy_http_request_parse_request(request,request->pstart1,request->pend1)) return -1;
					request->ppos++;
					request->pstart1=request->ppos+1;
					request->pend1=request->pstart1;
					request->status=PROXY_HTTP_REQUEST_PARSE_STATUS_HEADER;
				}
				else request->pend1++;
				break;
		}
		request->ppos++;
	}
	return 0;
}